import pickle
import warnings
from typing import Any, Tuple

import dask.array as da
import numpy as np
import pytest

from resource_backed_dask_array import (
    ResourceBackedDaskArray,
    resource_backed_dask_array,
)


# a *reusable* file context manager
class FileContext:
    FILE_OPEN = False
    OPEN_COUNT = 0

    @property
    def closed(self):
        return not self.FILE_OPEN

    def __enter__(self) -> "FileContext":
        if not self.FILE_OPEN:
            self.OPEN_COUNT += 1
        self.FILE_OPEN = True
        return self

    def __exit__(self, *args: Any) -> None:
        self.FILE_OPEN = False


@pytest.fixture
def dask_arr() -> da.Array:
    called = [0]
    ctx = FileContext()

    def get_chunk(block_id: Tuple[int, ...]) -> np.ndarray:
        if not ctx.FILE_OPEN:
            warnings.warn("You didn't open the file!")
        nonlocal called

        if not isinstance(block_id, np.ndarray):
            called[0] += 1
        return np.arange(100).reshape(10, 10)[np.newaxis, np.newaxis]

    d = da.map_blocks(get_chunk, chunks=((1,) * 10, (1,) * 10, 10, 10), dtype=float)
    d.called = called
    d.ctx = ctx
    return d


@pytest.fixture
def wrapper(dask_arr: da.Array) -> ResourceBackedDaskArray:
    arr = ResourceBackedDaskArray.from_array(dask_arr, dask_arr.ctx)
    arr.called = dask_arr.called
    arr.ctx = dask_arr.ctx
    return arr


def test_array(dask_arr: da.Array) -> None:
    with pytest.warns(UserWarning):
        dask_arr.compute()

    assert dask_arr.ctx.OPEN_COUNT == 0

    with dask_arr.ctx:
        assert dask_arr.compute().shape == (10, 10, 10, 10)

    assert dask_arr.ctx.OPEN_COUNT == 1


def test_from_rbda(wrapper: ResourceBackedDaskArray) -> None:
    assert wrapper.from_array(wrapper, wrapper._context) is wrapper


def test_wrapper_compute(wrapper: ResourceBackedDaskArray) -> None:
    ary = wrapper.compute()
    assert isinstance(ary, np.ndarray)
    assert ary.shape == (10, 10, 10, 10)
    assert wrapper.called[0] == 100


def test_wrapper_asarray(wrapper: ResourceBackedDaskArray) -> None:
    assert wrapper.ctx.OPEN_COUNT == 0
    ary: np.ndarray = np.asarray(wrapper)
    assert isinstance(ary, np.ndarray)
    assert ary.shape == (10, 10, 10, 10)
    assert wrapper.ctx.OPEN_COUNT == 1
    assert wrapper.called[0] == 100


def test_wrapper_getitem(dask_arr: da.Array, wrapper: ResourceBackedDaskArray) -> None:
    dask_arr.ctx.FILE_OPEN = True
    a = dask_arr[0, 1:3]
    b = wrapper[0, 1:3]
    assert isinstance(a, da.Array)
    assert isinstance(b, ResourceBackedDaskArray)
    np.testing.assert_array_equal(a.compute(), b.compute())


def test_wrapper_methods(dask_arr: da.Array, wrapper: ResourceBackedDaskArray) -> None:
    assert isinstance(wrapper, ResourceBackedDaskArray)
    dmean = wrapper.mean()
    assert isinstance(dmean, ResourceBackedDaskArray)
    assert isinstance(dmean.compute(), float)
    with pytest.warns(UserWarning):
        assert dmean.compute() == dask_arr.mean().compute()

    # non array-returning methods don't return ResourceBackedDaskArray
    assert isinstance(wrapper.to_svg(), str)


def test_wrapper_ufunc(dask_arr: da.Array, wrapper: ResourceBackedDaskArray) -> None:
    amean = np.mean(dask_arr)
    pmean = np.mean(wrapper)
    assert isinstance(amean, da.Array)
    assert isinstance(pmean, ResourceBackedDaskArray)
    dask_arr.ctx.FILE_OPEN = True
    assert amean.compute() == pmean.compute()


def test_wrapper_repr(dask_arr: da.Array, wrapper: ResourceBackedDaskArray) -> None:
    assert repr(dask_arr) == repr(wrapper)
    assert repr(dask_arr.mean) == repr(wrapper.mean)


def test_pickle():
    dsk = da.random.random((4, 4))
    ctx = FileContext()
    ctx.FILE_OPEN = True
    rdba = resource_backed_dask_array(dsk, ctx)
    assert not rdba._context.closed
    pda = pickle.dumps(dsk)
    prbda = pickle.dumps(rdba)
    assert isinstance(pda, bytes)
    assert isinstance(prbda, bytes)
    unpickled_rbda = pickle.loads(prbda)
    assert np.array_equal(pickle.loads(pda).compute(), dsk.compute())
    assert np.array_equal(unpickled_rbda.compute(), rdba.compute())
    assert unpickled_rbda._context.closed
